// Use this sensor to measure KWH and Watt of your house meeter
// You need to set the correct pulsefactor of your meeter (blinks per KWH).
// The sensor starts by fetching current KWH value from gateway.
// Reports both KWH and Watt back to gateway.
//
// Unfortunately millis() won't increment when the Arduino is in 
// sleepmode. So we cannot make this sensor sleep if we also want 
// to calculate/report watt-number.

#include <SPI.h>
#include <MySensor.h>  

#define DIGITAL_INPUT_SENSOR 3  // The digital input you attached your light sensor.  (Only 2 and 3 generates interrupt!)
#define PULSE_FACTOR 1000       // Nummber of blinks per KWH of your meeter
#define SLEEP_MODE true        // Watt-value can only be reported when sleep mode is false.
#define MAX_WATT 10000          // Max watt value to report. This filetrs outliers.
#define INTERRUPT DIGITAL_INPUT_SENSOR-2 // Usually the interrupt = pin -2 (on uno/nano anyway)

// Id of the sensor child
#define CHILD_ID_CNT 1
#define CHILD_ID_DEV 254

unsigned long SEND_FREQUENCY = 300000; // Minimum time between send (in milliseconds). We don't wnat to spam the gateway.
MySensor gw;
double ppwh = ((double)PULSE_FACTOR)/1000; // Pulses per watt hour
boolean pcReceived = false;
volatile unsigned long pulseCount = 0;   
volatile unsigned long lastBlink = 0;
volatile unsigned long watt = 0;
unsigned long oldPulseCount = 0;   
unsigned long oldWatt = 0;
double oldKwh;
double oldTemp = 0;
float oldBatteryV = 0;
unsigned long lastSend;

MyMessage CntWattMsg(CHILD_ID_CNT, V_WATT);
MyMessage CntKwhMsg(CHILD_ID_CNT,  V_KWH);
MyMessage pcMsg(CHILD_ID_CNT,      V_VAR1);
MyMessage pcReqMsg(CHILD_ID_CNT,   V_VAR2);

MyMessage TempMsg(CHILD_ID_DEV, V_TEMP);
MyMessage BattMsg(CHILD_ID_DEV, V_VOLTAGE);

long readVcc() {
  long result;
  // Read 1.1V reference against AVcc
  ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
  delay(2); // Wait for Vref to settle
  ADCSRA |= _BV(ADSC); // Convert
  while (bit_is_set(ADCSRA,ADSC));
  result = ADCL;
  result |= ADCH<<8;
  result = 1126400L / result; // Back-calculate AVcc in mV
  return result;
}

double GetTemp(void)
{
  unsigned int wADC;
  double t;

  // The internal temperature has to be used
  // with the internal reference of 1.1V.
  // Channel 8 can not be selected with
  // the analogRead function yet.

  // Set the internal reference and mux.
  ADMUX = (_BV(REFS1) | _BV(REFS0) | _BV(MUX3));
  ADCSRA |= _BV(ADEN);  // enable the ADC

  delay(20);            // wait for voltages to become stable.

  ADCSRA |= _BV(ADSC);  // Start the ADC

  // Detect end-of-conversion
  while (bit_is_set(ADCSRA,ADSC));

  // Reading register "ADCW" takes care of how to read ADCL and ADCH.
  wADC = ADCW;

  // The offset of 324.31 could be wrong. It is just an indication.
  t = (wADC - 324.31 ) / 1.22;

  // The returned temperature is in degrees Celcius.
  return (t);
}

void SendDevInfo()
{
  //========= Temperature =============
  double fTemp = GetTemp();
  
  if (oldTemp != fTemp){
    Serial.print("Temp:");
    Serial.println(fTemp);
  
    gw.send(TempMsg.set(fTemp, 2));
      
    oldTemp = fTemp;
  }

  //========= Battery =============  
  float batteryV  = readVcc() * 0.001;

  Serial.print("BatV:");
  Serial.println(batteryV);
  
  gw.send(BattMsg.set(batteryV, 2));
}

void setup()  
{  
  wdt_disable();
  
  // Disable the ADC by setting the ADEN bit (bit 7) to zero.
  ADCSRA = ADCSRA & B01111111;
  
  // Disable the analog comparator by setting the ACD bit
  // (bit 7) to one.
  ACSR = B10000000;   
  
  gw.begin(incomingMessage);

  // Send the sketch version information to the gateway and Controller
  gw.sendSketchInfo("Natural GAS Meter", "1.0");

  // Register this device as power sensor
  gw.present(CHILD_ID_CNT, S_POWER);
  gw.present(CHILD_ID_DEV, S_TEMP);
  gw.present(CHILD_ID_DEV, V_VOLTAGE);  

  // Fetch last known pulse count value from gw
  //gw.request(CHILD_ID_CNT, V_VAR1);
  gw.send(pcReqMsg.set("REQ"));
 
  attachInterrupt(INTERRUPT, onPulse, RISING);
  
  wdt_enable(WDTO_8S);  
  
  lastSend=millis();
}

void loop()     
{ 
  wdt_reset();
  
  // By calling process() you route messages in the background
  gw.process();

  // Send feq
  // Only send values at a maximum frequency or woken up from sleep
  unsigned long now = millis();
  bool sendTime = now - lastSend > SEND_FREQUENCY;

  //========= Cnt =============   
  if (pcReceived && (SLEEP_MODE || sendTime)) {
    SendDevInfo();
    
    // New watt value has been calculated  
    if (!SLEEP_MODE && watt != oldWatt) {
      // Check that we dont get unresonable large watt value. 
      // could hapen when long wraps or false interrupt triggered
      if (watt<((unsigned long)MAX_WATT)) {
        gw.send(CntWattMsg.set(watt));  // Send watt value to gw 
      }  
      Serial.print("Watt:");
      Serial.println(watt);
      oldWatt = watt;
    }
  
    // Pulse cout has changed
    if (pulseCount != oldPulseCount) {
      gw.send(pcMsg.set(pulseCount));  // Send pulse count value to gw 
      double kwh = ((double)pulseCount/((double)PULSE_FACTOR));     
      oldPulseCount = pulseCount;
      if (kwh != oldKwh) {
        gw.send(CntKwhMsg.set(kwh, 4));  // Send kwh value to gw 
        oldKwh = kwh;
      }
    }       
    
    lastSend = now;    
  } else if (sendTime && !pcReceived) {
    // No count received. Try requesting it again
    //gw.request(CHILD_ID_CNT, V_VAR1);
    gw.send(pcReqMsg.set("REQ"));
    
    lastSend=now;       
  }
  
  //========= Sleep ================
  if (SLEEP_MODE && pcReceived) {
    Serial.println("sleep");
    gw.sleep(SEND_FREQUENCY);
  }
}

void incomingMessage(const MyMessage &message) {
  if (message.type==V_VAR1) {  
    pulseCount = oldPulseCount = message.getLong();
    Serial.print("Received last pulse count from gw:");
    Serial.println(pulseCount);
    pcReceived = true;
  }
}

void onPulse()     
{ 
  if (!SLEEP_MODE) {
    unsigned long newBlink = micros();  
    unsigned long interval = newBlink-lastBlink;
    if (interval<10000L) { // Sometimes we get interrupt on RISING
      return;
    }
    watt = (3600000000.0 /interval) / ppwh;
    lastBlink = newBlink;
  } 
  pulseCount++;
}



